=begin
    mo.rb - A simple class for reading GNU MO file.
    Copyright (C) 2001  Masahiro Sakai

        Masahiro Sakai                  <s01397ms@sfc.keio.ac.jp>

    You can redistribute this file and/or modify it under the same term
    of Ruby.  License of Ruby is included with Ruby distribution in
    the file "README".
=end

require 'delegate'

class MOFile < DelegateClass(Hash)
  class InvalidFormat < RuntimeError; end;

  Header = Struct.new(:magic,
                      :revision,
                      :nstrings,
                      :orig_table_offset,
                      :translated_table_offset,
                      :hash_table_size,
                      :hash_table_offset)

  MAGIC_BIG_ENDIAN    = "\x95\x04\x12\xde"
  MAGIC_LITTLE_ENDIAN = "\xde\x12\x04\x95"

  def self.open(arg = nil)
    result = self.new
    case arg
    when String
      result.load_from_file(arg)
    when IO
      result.load_from_stream(arg)
    end
    result
  end

  def initialize
    @hash = Hash.new()
    @little_endian = false
    super(@hash)
  end

  def load_from_stream(io)
    magic = io.read(4)
    case magic
    when MAGIC_BIG_ENDIAN
      @little_endian = false
    when MAGIC_LITTLE_ENDIAN
      @little_endian = true
    else
      raise InvalidFormat.new(sprintf("Unknown signature %s", magic.dump))
    end

    header = Header.new(magic, *(io.read(4 * 6).unpack(@little_endian ? 'V6' : 'N6')))
    raise InvalidFormat.new(sprintf("file format revision %d isn't supported", header.revision)) if header.revision > 0

    io.pos = header.orig_table_offset
    orig_table_data = io.read((4 * 2) * header.nstrings).unpack(@little_endian ? 'V*' : 'N*')

    io.pos = header.translated_table_offset
    trans_table_data = io.read((4 * 2) * header.nstrings).unpack(@little_endian ? 'V*' : 'N*')

    original_strings = Array.new(header.nstrings)
    for i in 0...header.nstrings
      io.pos = orig_table_data[i * 2 + 1]
      original_strings[i] = io.read(orig_table_data[i * 2 + 0])
    end

    @hash.clear
    for i in 0...header.nstrings
      io.pos = trans_table_data[i * 2 + 1]
      @hash[original_strings[i]] = io.read(trans_table_data[i * 2 + 0])
    end

    self
  end

  def save_to_stream(io)
    header_size = 4 * 7
    table_size  = 4 * 2 * @hash.size

    header = Header.new(
      @little_endian ? MAGIC_LITTLE_ENDIAN : MAGIC_BIG_ENDIAN, # magic
      0,                            # revision
      @hash.size,                   # nstrings
      header_size,                  # orig_table_offset
      header_size + table_size,     # translated_table_offset
      header_size + table_size * 2, # hash_table_offset
      0                             # hash_table_size
    )
    io.write(header.to_a.pack('a4' + (@little_endian ? 'V6' : 'N6')))

    ary = @hash.to_a
    ary.sort!{|a, b| a[0] <=> b[0]} # sort by original string

    pos = header.hash_table_offset + header.hash_table_size

    orig_table_data = Array.new()
    ary.each{|item, _|
      orig_table_data.push(item.size)
      orig_table_data.push(pos)
      pos += item.size + 1 # +1 is <NUL>
    }
    io.write(orig_table_data.pack(@little_endian ? 'V*' : 'N*'))

    trans_table_data = Array.new()
    ary.each{|_, item|
      trans_table_data.push(item.size)
      trans_table_data.push(pos)
      pos += item.size + 1 # +1 is <NUL>
    }
    io.write(trans_table_data.pack(@little_endian ? 'V*' : 'N*'))

    ary.each{|item, _| io.write(item); io.write("\0") }
    ary.each{|_, item| io.write(item); io.write("\0") }

    self
  end

  def load_from_file(filename)
    File.open(filename, 'rb'){|f| load_from_stream(f)}
  end

  def save_to_file(filename)
    File.open(filename, 'wb'){|f| save_to_stream(f)}
  end

  def charset
    header = self['']
    if header
      header.each_line{|line|
        if /^Content-Type:/i =~ line and /charset=((?:\w|-)+)/i =~ line
          return $1
        end
      }
    end
    return nil
  end

  attr_accessor :little_endian
end


# Test

if $0 == __FILE__
  if (ARGV.include? "-h") or (ARGV.include? "--help")
    STDERR.puts("mo.rb [filename.mo ...]")
    exit
  end

  ARGV.each{ |item|
    mo = MOFile.open(item)
    puts("------------------------------------------------------------------")
    puts("charset = " + mo.charset) if mo.charset
    puts("------------------------------------------------------------------")
    mo.to_hash.each{
      |key, value|
      puts("original message = \"#{key}\"")
      puts("translated message = \"#{value}\"")
      puts("--------------------------------------------------------------------")
    }
  }
end
